#######################################################################################################################################################################################################
#
# Changes to the method
# General
#     2022-Feb-01: add method for ClimaCache.LogisticVC
#     2022-Feb-01: add documentation
#
#######################################################################################################################################################################################################
"""

    relative_hydraulic_conductance(vc::ComplexVC{FT}, p_25::FT) where {FT<:AbstractFloat}

Return the hydraulic conductance ralative to maximum at reference temperature, given
- `vc` `ClimaCache.ComplexVC` type vulnerability curve
- `p_25` Equivalent xylem water pressu

---
# Examples
```julia
vc = ComplexVC{Float64}([0.3,0.3,0.4], [LogisticVC{FT}(2,2), PowerVC{FT}(2,2), WeibullVC{FT}(2,2)]);
kr = relative_hydraulic_conductance(vs4, FT(-2.0));
```
"""
relative_hydraulic_conductance(vc::ComplexVC{FT}, p_25::FT) where {FT<:AbstractFloat} = (
    @unpack PS, VCS = vc;
    @assert sum(PS) ≈ 1 "Probabilities of VCs must sum up to 1!";
    @assert length(PS) == length(VCS) "Lengths of VC curves and probabilities must equal!";

    _kr::FT = 0;
    for _i in eachindex(PS)
        _kr += relative_hydraulic_conductance(VCS[_i], p_25) * PS[_i];
    end;

    return _kr
);


#######################################################################################################################################################################################################
#
# Changes to the method
# General
#     2022-Feb-01: add method for ClimaCache.LogisticVC
#     2022-Feb-01: add documentation
#
#######################################################################################################################################################################################################
"""

    relative_hydraulic_conductance(vc::LogisticVC{FT}, p_25::FT) where {FT<:AbstractFloat}

Return the hydraulic conductance ralative to maximum at reference temperature, given
- `vc` `ClimaCache.LogisticVC` type vulnerability curve
- `p_25` Equivalent xylem water pressure at 298.15 K in `[MPa]` (surface tension correction made)

---
# Examples
```julia
vc = LogisticVC{Float64}(2, 2);
kr = relative_hydraulic_conductance(vc, -2.0);
```
"""
relative_hydraulic_conductance(vc::LogisticVC{FT}, p_25::FT) where {FT<:AbstractFloat} = (
    if p_25 >= 0
        return FT(1)
    end;

    @unpack A, B = vc;

    return max( eps(FT), (1 - 1/(1 + A * exp(B * p_25))) * (A+1)/A )
);


#######################################################################################################################################################################################################
#
# Changes to the method
# General
#     2022-Feb-01: add method for ClimaCache.PowerVC
#     2022-Feb-01: add documentation
#
#######################################################################################################################################################################################################
"""

    relative_hydraulic_conductance(vc::PowerVC{FT}, p_25::FT) where {FT<:AbstractFloat}

Return the hydraulic conductance ralative to maximum at reference temperature, given
- `vc` `ClimaCache.PowerVC` type vulnerability curve
- `p_25` Equivalent xylem water pressure at 298.15 K in `[MPa]` (surface tension correction made)

---
# Examples
```julia
vc = PowerVC{Float64}(2, 2);
kr = relative_hydraulic_conductance(vc, -2.0);
```
"""
relative_hydraulic_conductance(vc::PowerVC{FT}, p_25::FT) where {FT<:AbstractFloat} = (
    if p_25>=0
        return FT(1)
    end;

    @unpack A, B = vc;

    return max( eps(FT), 1 / (1 + A*(-p_25)^B) )
);


#######################################################################################################################################################################################################
#
# Changes to the method
# General
#     2022-Feb-01: add method for ClimaCache.WeibullVC
#     2022-Feb-01: add documentation
#
#######################################################################################################################################################################################################
"""

    relative_hydraulic_conductance(vc::WeibullVC{FT}, p_25::FT) where {FT<:AbstractFloat}

Return the hydraulic conductance ralative to maximum at reference temperature, given
- `vc` `ClimaCache.WeibullVC` type vulnerability curve
- `p_25` Equivalent xylem water pressure at 298.15 K in `[MPa]` (surface tension correction made)

---
# Examples
```julia
vc = WeibullVC{Float64}(2, 2);
kr = relative_hydraulic_conductance(vc, -2.0);
```
"""
relative_hydraulic_conductance(vc::WeibullVC{FT}, p_25::FT) where {FT<:AbstractFloat} = (
    if p_25>=0
        return FT(1)
    end;

    @unpack B, C = vc;

    return max( eps(FT), exp( -1 * (-p_25 / B) ^ C ) )
);
